package cz.zcu.ups.arcaoftpd;

import java.net.InetSocketAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.ArrayList;
import java.util.Date;

import cz.zcu.ups.arcaoftpd.client.ClientHandler;
import cz.zcu.ups.arcaoftpd.client.ClientParameters;
import cz.zcu.ups.arcaoftpd.users.Users;
import cz.zcu.ups.arcaoftpd.util.Logger;

/**
 * Hlavni trida programu
 * @author Arcao
 */
public class ArcaoFTPd {
	private static Users users;
	private static Logger logger;
	private ServerSocket serverSocket;
	private ArrayList<ClientHandler> chList;
	/**
	 * Vychozi port na kterem bude ftp server naslouchat
	 */
	public static int DEFAULT_PORT = 2121;

	/**
	 * Sposteci trida programu
	 * @param args			parametry programu
	 */
	public static void main(String[] args) {
		int port = DEFAULT_PORT;

		System.out.println("ArcaoFTPd verze 1.0 - FTP Server");
		System.out.println("Autor: Martin \"Arcao\" Sloup; UID: A04372; E-mail: msloup@students.zcu.cz");
		System.out.println();
		System.out.println("Tento program vznikl jako semestralni prace z predmetu KIV/UPS - Uvod do pocitacovych siti");

		if (args.length == 1) {
			try {
				port = Integer.parseInt(args[0]);
				if (port < 1 || port > 65535) throw new Exception();
			} catch(Exception e) {
				System.out.println("Nespravne cislo portu! Bude pouzit vychozi port " + DEFAULT_PORT);
			}
		} else {
			System.out.println("Nebyl zadan jako prvni parametr port, na kterem se bude naslouchat,");
			System.out.println("proto se pouzije vychozi port " + DEFAULT_PORT);
		}

		ArcaoFTPd app = new ArcaoFTPd(port);
		app.serverLoop();
	}

	/**
	 * Konstruktor hlavni tridy
	 * @param port			port na kterem bude naslouchat ftp server
	 */
	public ArcaoFTPd(int port) {
		try {
			serverSocket = new ServerSocket();
			chList = new ArrayList<ClientHandler>();

			//nabindujeme ftp server na zadanem portu
			serverSocket.bind(new InetSocketAddress(port));
			System.out.println("FTP Server spusten na portu " + port);

			users = new Users("users.xml");
			logger = new Logger("access.log", "error.log");

			logger.println("--------------------------------------------------------------------------------");
			logger.println("I: " + new Date().toString() +  ": ArcaoFTPd starting...");
			logger.println("I: " + new Date().toString() +  ": Port " + port + " binded...");

			Runtime.getRuntime().addShutdownHook(new ShutdownThread(this));
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * Cyklus, ve kterem server ceka na pripojeni dalsich klientu
	 */
	public void serverLoop() {
		try {
			while(!serverSocket.isClosed() && serverSocket.isBound()) {
				Socket s = serverSocket.accept();

				ClientHandler ch = new ClientHandler(new ClientParameters(s));
				ch.start();
				chList.add(ch);
			}
		} catch (Exception e) {
			//e.printStackTrace();
		}
	}

	/**
	 * Vrati objekt s informacemi o uzivatelych
	 * @return			objekt s informacemi o uzivatelych
	 */
	public static Users getUsers() {
		return users;
	}

	/**
	 * Vrati objekt loggeru
	 * @return			objekt loggeru
	 */
	public static Logger getLogger() {
		return logger;
	}

	/**
	 * Ukonceni vsech zdroju
	 */
	public void dispose() {
		try {
			System.out.println("Ukoncuji ArcaoFTPd...");

			for (ClientHandler item : chList)
				item.dispose();
			chList = null;

			logger.println("I: " + new Date().toString() +  ": ArcaoFTPd stopping...");
			serverSocket.close();
			logger = null;
		} catch(Exception e) {
		}
	}

	/**
	 * Trida pro hlidani nasilneho ukonceni aplikace
	 */
	class ShutdownThread extends Thread {
		  private ArcaoFTPd app = null;

		  /**
		   * Konstroktor tridy
		   * @param app			hlavni trida apliace
		   */
		  public ShutdownThread(ArcaoFTPd app) {
		    super();
		    this.app = app;
		  }

		  public void run() {
			app.dispose();
			app = null;
		  }
		}

}
