package cz.zcu.ups.arcaoftpd.client;

import java.io.BufferedReader;
import java.io.File;
import java.io.InputStreamReader;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.SocketAddress;
import java.text.SimpleDateFormat;
import java.util.Date;

import cz.zcu.ups.arcaoftpd.ArcaoFTPd;
import cz.zcu.ups.arcaoftpd.connection.ActiveConnection;
import cz.zcu.ups.arcaoftpd.connection.PassiveConnection;
import cz.zcu.ups.arcaoftpd.dataaction.ListDataAction;
import cz.zcu.ups.arcaoftpd.dataaction.NlstDataAction;
import cz.zcu.ups.arcaoftpd.dataaction.RetriveDataAction;
import cz.zcu.ups.arcaoftpd.dataaction.StoreDataAction;
import cz.zcu.ups.arcaoftpd.users.User;

/**
 * Trida starajici se o pripojeneho klienta, obsluhuje vsechny poslane prikazy
 * @author Arcao
 */
public class ClientHandler extends Thread {
	private ClientParameters clientParameters;
	private BufferedReader commandStream;

	/**
	 * Konstruktor tridy
	 * @param clientParameters		informace o pripojenem klientovi
	 */
	public ClientHandler(ClientParameters clientParameters) {
		this.clientParameters = clientParameters;
	}

	/**
	 * Hlavni fuknkce vlakna, stara se o nacitani vstupu ze soketu a pripadne vykonani prikazu
	 */
	public void run() {
		String line;
		try {
			ArcaoFTPd.getLogger().write(("O: " + new Date().toString() + "; " + clientParameters.getSocket().getRemoteSocketAddress().toString() +": Connection opened & accepted...\n").getBytes("utf-8"));

			commandStream = new BufferedReader(new InputStreamReader(clientParameters.getSocket().getInputStream(), clientParameters.getEncoding()));

			clientParameters.sendMessage(220, "ArcaoFTPd ready...");

			while(!clientParameters.getClosed() & !clientParameters.getSocket().isClosed() && (line  = commandStream.readLine()) != null) {
					ArcaoFTPd.getLogger().write(("R: " + new Date().toString() + "; " + clientParameters.getSocket().getRemoteSocketAddress().toString() +": "+ line + "\n").getBytes("utf-8"));
					doCommand(line.trim());
			}
		} catch (Exception ex) {
		} finally {
			try {
				ArcaoFTPd.getLogger().write(("C: " + new Date().toString() + "; " + clientParameters.getSocket().getRemoteSocketAddress().toString() +": Connection closed...\n").getBytes("utf-8"));
				clientParameters.getSocket().close();
			} catch (Exception e) {
			}
		}
	}

	/**
	 * Znovu vytvori commandStream. Vola se po zmene kodovani (prikaz OPTS UTF8 ON/OFF)
	 */
	private void resetCommandEncoding() {
		try {
			commandStream = new BufferedReader(new InputStreamReader(clientParameters.getSocket().getInputStream(), clientParameters.getEncoding()));
		} catch (Exception ex) {
			ex.printStackTrace();
		}
	}

	/**
	 * Funkce vykonava prikazy FTP protokolu
	 * @param line		radka s prikazem a parametry
	 */
	private void doCommand(String line) {
		String command, parameter, filename, path;
		File file;

		try {
			String[] cmdParseArr = line.split(" ", 2);
			command = cmdParseArr[0].toUpperCase();
			parameter = "";
			if (cmdParseArr.length == 2) parameter = cmdParseArr[1].trim();
			//OutputStream out = clientParameters.getSocket().getOutputStream();

			//USER <username>
			if (command.equals("USER")) {
				clientParameters.setUsername(parameter);
				clientParameters.sendMessage(331, "User name okay, need password.");
			//PASS <password>
			} else if (command.equals("PASS")) {
				// Neni nastaven uzivatel => Chyba!
				if (clientParameters.getUsername().length() == 0) {
					clientParameters.sendMessage(503, "Bad sequence of commands.");
					clientParameters.setClosed(true);
					return;
				}

				try {
					User user = ArcaoFTPd.getUsers().getUser(clientParameters.getUsername());
					if (user == null) throw new Exception();

					if (user.isPasswordReq() && !user.getPassword().equals(parameter)) throw new Exception();

					clientParameters.setUser(user);
					clientParameters.sendMessage(230, "User logged in, proceed");
				} catch (Exception e) {
					clientParameters.sendMessage(530, "Not logged in.");
					//clientParameters.setClosed(true);
				}
			} else if (clientParameters.getLogged()) {
				//TYPE (A|A N|I|L 8)
				if (command.equals("TYPE")) {
					if (parameter.equalsIgnoreCase("A") || parameter.equalsIgnoreCase("A N")) {
						clientParameters.setTransferMode(TransferMode.Ascii);
						clientParameters.sendMessage(200, "Command okay.");
					} else if (parameter.equalsIgnoreCase("I") || parameter.equalsIgnoreCase("L 8")) {
						clientParameters.setTransferMode(TransferMode.Binary);
						clientParameters.sendMessage(200, "Command okay");
					} else {
						clientParameters.sendMessage(504, "Command not implemented for that parameter");
					}
				//STRU F
				} else if (command.equals("STRU")) {
					if (parameter.equalsIgnoreCase("F")) {
						clientParameters.sendMessage(200, "Command okay.");
					} else {
						clientParameters.sendMessage(504, "Command not implemented for that parameter.");
					}
				//MODE S
				} else if (command.equals("MODE")) {
					if (parameter.equalsIgnoreCase("S")) {
						clientParameters.sendMessage(200, "Command okay.");
					} else {
						clientParameters.sendMessage(504, "Command not implemented for that parameter.");
					}
				//[X]CWD <path>
				} else if (command.equals("CWD") || command.equals("XCWD")) {
					try {
						String newPath = clientParameters.getPath() + "/" + parameter;
						if (parameter.charAt(0) == '/') newPath = parameter;
						if (newPath.length() > 1 && newPath.charAt(newPath.length() - 1) == '/') newPath = newPath.substring(0, newPath.length() - 1);

						if (newPath.length() > 1 && newPath.substring(0,2).equals("//")) newPath = newPath.substring(1);

						if ((newPath + "/").indexOf("/../") != -1 ||
							(newPath + "/").indexOf("/./") != -1 ||
							!clientParameters.getUser().exists(newPath) ||
							!clientParameters.getUser().isPathListable(newPath)) throw new Exception();

						clientParameters.setPath(newPath);

						clientParameters.sendMessage(200, "Command okay.");
					} catch (Exception e) {
						clientParameters.sendMessage(550, "%s: No such file or directory.", parameter);
					}
				//[X]PWD
				} else if (command.equals("PWD") || command.equals("XPWD")) {
					clientParameters.sendMessage(257, "\"%s\" is current directory.", clientParameters.getPath());
				//(CDUP|XCUP)
				} else if (command.equals("CDUP") || command.equals("XCUP")) {
					try {
						if (clientParameters.getPath().equals("/")) throw new Exception();

						if (clientParameters.getPath().lastIndexOf('/') > 0) {
							clientParameters.setPath(clientParameters.getPath().substring(0, clientParameters.getPath().lastIndexOf('/')));
						} else {
							clientParameters.setPath("/");
						}
						clientParameters.sendMessage(200, "Okay.");
					} catch (Exception e) {
						clientParameters.sendMessage(550, "Requested action not taken.");
					}
				//PASV
				} else if (command.equals("PASV")) {
					clientParameters.setDataConnection(new PassiveConnection(clientParameters));
				//RETR <filename>
				} else if (command.equals("RETR")) {
					try {
						if (clientParameters.getDataConnection() == null) {
							clientParameters.sendMessage(550, "No port specified.");
							return;
						}

						path = clientParameters.getPath() + "/" + parameter;
						if (parameter.charAt(0) == '/') path = parameter;
						if (path.length() > 1 && path.substring(0,2).equals("//")) path = path.substring(1);

						// zkontrolujeme, ze je pravo na listovani
						if (!clientParameters.getUser().exists(path) ||
							!clientParameters.getUser().isFile(path) ||
							!clientParameters.getUser().isPathReadable(path)) {
							clientParameters.sendMessage(550, "Cannot RETR. No permission.");
							clientParameters.setDataConnection(null);
							return;
						}

						clientParameters.getDataConnection().setAction(new RetriveDataAction(path));
						clientParameters.getDataConnection().connect();
					} catch (Exception e) {
						clientParameters.setDataConnection(null);
						clientParameters.sendMessage(451, "Requested action aborted: local error in processing.");
					}
				//REST <position>
				} else if (command.equals("REST")) {
					clientParameters.setFileSeekPosition(Integer.parseInt(parameter, 10));
					clientParameters.sendMessage(350, "Requested file action pending further information.");
				//PORT <ip1>,<ip2>,<ip3>,<ip4>,<port1>,<port2>
				} else if (command.equals("PORT")) {
					try {
						byte[] ip = new byte[4];
						int port;

						if (parameter.length() != 0) {

							String[] splittedPort = parameter.split(",");

							for (int i = 0; i < 4; i++)
								ip[i] = (byte) Integer.parseInt(splittedPort[i]);
							port = 256 * Integer.parseInt(splittedPort[4]) + Integer.parseInt(splittedPort[5]);
						} else {
							ip = clientParameters.getSocket().getInetAddress().getAddress();
							port = 20;
						}

						for (int i=0; i < 4; i++) {
							if (ip[i] != clientParameters.getSocket().getInetAddress().getAddress()[i]) throw new Exception();
						}

						SocketAddress portDestination = new InetSocketAddress(InetAddress.getByAddress(ip), port);
						clientParameters.setDataConnection(new ActiveConnection(clientParameters, portDestination));

						clientParameters.sendMessage(200, "Port command successful.");
					} catch (Exception e) {
						e.printStackTrace();
						clientParameters.sendMessage(550, "Requested action not taken. ");
					}
				//LIST [<parameter> ][<path>]
				} else if (command.equals("LIST")) {
					try {
						if (clientParameters.getDataConnection() == null) {
							clientParameters.sendMessage(550, "No port specified.");
							return;
						}
						// zrusime seek position, v LIST se na seek nehledi
						clientParameters.setFileSeekPosition(0);

						if (parameter.length() > 0) {
							//odstranime parametr pro listovani
							parameter = parameter.replaceAll("^-[LlaRT]+\\s*", "");
						}
						parameter = parameter.trim();

						//nastavime cestu
						path = clientParameters.getPath();
						//pokud parametr obsahuje cestu, pak ho pridame, nebo nsatavime pokud je parametr absolutni cesta
						if (parameter.length() > 0) {
							path = path + "/" + parameter;
							if (parameter.charAt(0) == '/') path = parameter;
							if (path.length() > 1 && path.substring(0,2).equals("//")) path = path.substring(1);
						}

						// zkontrolujeme, ze je pravo na listovani
						if (!clientParameters.getUser().exists(path) ||
							!clientParameters.getUser().isPathListable(path)) {
							clientParameters.sendMessage(550, "Cannot LIST. No permission.");
							clientParameters.setDataConnection(null);
							return;
						}

						clientParameters.getDataConnection().setAction(new ListDataAction(path));
						clientParameters.getDataConnection().connect();
					} catch (Exception e) {
						clientParameters.setDataConnection(null);
						clientParameters.sendMessage(451, "Requested action aborted: local error in processing.");
					}
				//NLST [<path>]
				} else if (command.equals("NLST")) {
					try {
						if (clientParameters.getDataConnection() == null) {
							clientParameters.sendMessage(550, "No port specified.");
							return;
						}
						// zrusime seek position, v NLST se na seek nehledi
						clientParameters.setFileSeekPosition(0);

						//nastavime cestu
						path = clientParameters.getPath();
						//pokud parametr obsahuje cestu, pak ho pridame, nebo nsatavime pokud je parametr absolutni cesta
						if (parameter.length() > 0) {
							path = path + "/" + parameter;
							if (parameter.charAt(0) == '/') path = parameter;
							if (path.length() > 1 && path.substring(0,2).equals("//")) path = path.substring(1);
						}

						// zkontrolujeme, ze je pravo na listovani
						if (!clientParameters.getUser().exists(path) ||
							!clientParameters.getUser().isDirecory(path) ||
							!clientParameters.getUser().isPathListable(path)) {
							clientParameters.sendMessage(550, "Cannot NLST. No permission.");
							clientParameters.setDataConnection(null);
							return;
						}

						clientParameters.getDataConnection().setAction(new NlstDataAction(path));
						clientParameters.getDataConnection().connect();
					} catch (Exception e) {
						clientParameters.setDataConnection(null);
						clientParameters.sendMessage(451, "Requested action aborted: local error in processing.");
					}
				//QUIT
				} else if (command.equals("QUIT")) {
					clientParameters.sendMessage(221, "Bye.");
					clientParameters.setClosed(true);
				//SYST
				} else if (command.equals("SYST")) {
					clientParameters.sendMessage(215, "UNIX Type: L8");
				//STAT
				} else if (command.equals("STAT")) {
					clientParameters.sendMessage(504, "Command not implemented for that parameter.");
				//NOOP
				} else if (command.equals("NOOP")) {
					//TODO dodelat vynulovani casovace
					clientParameters.sendMessage(200, "Command okay.");
				//STOR <filename>, APPE <filename>
				} else if (command.equals("STOR") || command.equals("APPE")) {
					try {
						if (clientParameters.getDataConnection() == null) {
							clientParameters.sendMessage(550, "No port specified.");
							return;
						}

						path = clientParameters.getPath() + "/" + parameter;
						if (parameter.charAt(0) == '/') path = parameter;
						if (path.length() > 1 && path.substring(0,2).equals("//")) path = path.substring(1);

						// zkontrolujeme, ze je pravo na listovani
						if ((path + "/").indexOf("/../") != -1 ||
							(path + "/").indexOf("/./") != -1 ||
							!clientParameters.getUser().isPathWriteable(path) ||
							(command.equals("APPE") && !clientParameters.getUser().isPathFileAppendable(path))) {
							clientParameters.setDataConnection(null);
							clientParameters.sendMessage(550, "Cannot %s. No permission.", command);
							return;
						}

						clientParameters.getDataConnection().setAction(new StoreDataAction(path, command.equals("APPE")));
						clientParameters.getDataConnection().connect();
					} catch (Exception e) {
						clientParameters.setDataConnection(null);
						clientParameters.sendMessage(451, "Requested action aborted: local error in processing.");
					}
				//ALLO
				} else if (command.equals("ALLO")) {
					clientParameters.sendMessage(202, "Command not implemented, superfluous at this site.");
				//[X]MKD <dirname>
				} else if (command.equals("MKD") || command.equals("XMKD")) {
					try {
						path = clientParameters.getPath() + "/" + parameter;
						if (parameter.charAt(0) == '/') path = parameter;
						if (path.substring(0,2).equals("//")) path = path.substring(1);

						if (clientParameters.getUser().exists(path) ||
							path.indexOf("/../") != -1 ||
							path.indexOf("/./") != -1 ||
							!clientParameters.getUser().isPathDirectoryMakeable(path)) {
							clientParameters.sendMessage(550, "Cannot %s. No permission.", command);
							return;
						}

						file = clientParameters.getUser().translatePathToAbsoluteFile(path);
						if (!file.mkdirs()) throw new Exception();

						clientParameters.sendMessage(250, "Requested file action okay, completed.");
					} catch (Exception e) {
						clientParameters.sendMessage(550, "Requested action not taken.");
					}
				//[X]RMD <dirname>
				} else if (command.equals("RMD") || command.equals("XRMD")) {
					try {
						path = clientParameters.getPath() + "/" + parameter;
						if (parameter.charAt(0) == '/') path = parameter;
						if (path.length() > 1 && path.substring(0,2).equals("//")) path = path.substring(1);

						if (!clientParameters.getUser().exists(path) ||
							!clientParameters.getUser().isDirecory(path) ||
							path.indexOf("/../") != -1 ||
							path.indexOf("/./") != -1 ||
							!clientParameters.getUser().isPathDirectoryDeleteable(path)) {
							clientParameters.sendMessage(550, "Cannot %s. No permission.", command);
						}

						file = clientParameters.getUser().translatePathToAbsoluteFile(path);
						if (!file.delete()) throw new Exception();

						clientParameters.sendMessage(250, "Requested file action okay, completed.");
					} catch (Exception e) {
						clientParameters.sendMessage(550, "Requested action not taken.");
					}
				//DELE <filename>
				} else if (command.equals("DELE")) {
					try {
						filename = clientParameters.getPath() + "/" + parameter;
						if (parameter.charAt(0) == '/') filename = parameter;
						if (filename.length() > 1 && filename.substring(0,2).equals("//")) filename = filename.substring(1);

						if (!clientParameters.getUser().exists(filename) ||
							!clientParameters.getUser().isFile(filename) ||
							filename.indexOf("/../") != -1 ||
							filename.indexOf("/./") != -1 ||
							!clientParameters.getUser().isPathFileDeleteable(filename)) {
							clientParameters.sendMessage(550, "Cannot %s. No permission.", command);
						}

						file = clientParameters.getUser().translatePathToAbsoluteFile(filename);
						if (!file.delete()) throw new Exception();

						clientParameters.sendMessage(250, "Requested file action okay, completed.");
					} catch (Exception e) {
						clientParameters.sendMessage(550, "Requested action not taken.");
					}
				//RNFR <name>
				} else if (command.equals("RNFR")) {
					try {
						filename = clientParameters.getPath() + "/" + parameter;
						if (parameter.charAt(0) == '/') filename = parameter;
						if (filename.length() > 1 && filename.substring(0,2).equals("//")) filename = filename.substring(1);

						if (!clientParameters.getUser().exists(filename) ||
							!clientParameters.getUser().isFile(filename) ||
							filename.indexOf("/../") != -1 ||
							filename.indexOf("/./") != -1 ||
							!clientParameters.getUser().isPathWriteable(filename) ||
							!clientParameters.getUser().isPathReadable(filename)) {
							clientParameters.sendMessage(550, "Cannot %s. No permission.", command);
						}

						clientParameters.setRenameFromFilename(filename);
						clientParameters.sendMessage(350, "Requested file action pending further information.");
					} catch (Exception e) {
						clientParameters.sendMessage(550, "Requested action not taken.");
					}
				//RNTO <name>
				} else if (command.equals("RNTO")) {
					try {
						filename = clientParameters.getPath() + "/" + parameter;
						if (parameter.charAt(0) == '/') filename = parameter;
						if (filename.length() > 1 && filename.substring(0,2).equals("//")) filename = filename.substring(1);

						if (clientParameters.getUser().exists(filename) ||
							filename.indexOf("/../") != -1 ||
							filename.indexOf("/./") != -1 ||
							!clientParameters.getUser().isPathWriteable(filename)) {
							clientParameters.sendMessage(550, "Cannot %s. No permission.", command);
						}

						file = clientParameters.getUser().translatePathToAbsoluteFile(clientParameters.getRenameFromFilename());
						file.renameTo(clientParameters.getUser().translatePathToAbsoluteFile(filename));

						clientParameters.sendMessage(250, "Requested file action okay, completed.");
					} catch (Exception e) {
						clientParameters.sendMessage(550, "Requested action not taken.");
					}
				//CLNT <client>
				} else if (command.equals("CLNT")) {
					clientParameters.sendMessage(200, "Noted.");
				//SIZE <filename>
				} else if (command.equals("SIZE")) {
					try {
						filename = clientParameters.getPath() + "/" + parameter;
						if (parameter.charAt(0) == '/') filename = parameter;
						if (filename.length() > 1 && filename.substring(0,2).equals("//")) filename = filename.substring(1);

						if (!clientParameters.getUser().exists(filename) ||
							!clientParameters.getUser().isFile(filename)) {
							clientParameters.sendMessage(550, "Cannot %s. No permission.", command);
						}

						clientParameters.sendMessage(213, "%d", clientParameters.getUser().translatePathToAbsoluteFile(filename).length());
					} catch(Exception e) {
						clientParameters.sendMessage(550, "Requested action not taken.");
					}
				//MDTM [yyyymmddhhmmss] <name>
				} else if (command.equals("MDTM")) {
					try {
						Date mdtm = null;

						if (parameter.matches("^\\d{14}\\s.+")) {
							String[] mdtmParts = parameter.split(" ", 2);
							mdtm = new SimpleDateFormat("yyyyMMddHHmmss").parse(mdtmParts[0]);
							parameter = mdtmParts[1];
						}
						filename = clientParameters.getPath() + "/" + parameter;
						if (parameter.charAt(0) == '/') filename = parameter;
						if (filename.length() > 1 && filename.substring(0,2).equals("//")) filename = filename.substring(1);

						if (!clientParameters.getUser().exists(filename)) {
							clientParameters.sendMessage(550, "Cannot %s. File \"%s\" not found.", command, parameter);
						}

						if (mdtm == null) {
							clientParameters.sendMessage(213, "%tY%<tm%<td%<tH%<tM%<tS", new Date(clientParameters.getUser().translatePathToAbsoluteFile(filename).lastModified()));
						} else {
							if (!clientParameters.getUser().isPathWriteable(filename)) throw new Exception();
							clientParameters.getUser().translatePathToAbsoluteFile(filename).setLastModified(mdtm.getTime());
						}
					} catch(Exception e) {
						clientParameters.sendMessage(550, "Requested action not taken.");
					}
				//FEAT
				} else if (command.equals("FEAT")) {
					clientParameters.sendMessage(211, "Features:\nMDTM\nREST STREAM\nSIZE\nUTF8\nEnd");
				//OPTS UTF8 (ON|OFF)
				} else if (command.equals("OPTS")) {
					if (parameter.equalsIgnoreCase("utf8 on")) {
						clientParameters.setEncoding("utf-8");
						resetCommandEncoding();
						clientParameters.sendMessage(200, "UTF8 OPTS ON");
					} else if (parameter.equalsIgnoreCase("utf8 off")) {
						clientParameters.setEncoding("cp1250");
						resetCommandEncoding();
						clientParameters.sendMessage(200, "UTF8 OPTS OFF");
					} else {
						clientParameters.sendMessage(504, "Command not implemented for that parameter.");
					}
				//SITE CHMOD - kvuli spetne kompatibilite, Midnight Commander to spatne nese, ze neni naimplementovan
				} else if (command.equals("SITE")) {
					if (parameter.length() > 5 && parameter.toLowerCase().substring(0, 6).equals("chmod ")) {
						clientParameters.sendMessage(200, "Command not implemented yet, but ok.");
					} else {
						clientParameters.sendMessage(504, "Command not implemented for that parameter.");
					}
				} else {
					clientParameters.sendMessage(500, "Syntax error, command unrecognized.");
				}
			} else {
				clientParameters.sendMessage(500, "Syntax error, command unrecognized.");
			}
		} catch (Exception e) {
			// TODO Auto-generated catch block
			try {
				clientParameters.sendMessage(500, "Syntax error, command unrecognized.");
			} catch (Exception ex) {
			}
		}
	}

	/**
	 * Uvolni vsechny zdroje
	 */
	public void dispose() {
		try {
			if (clientParameters.getDataConnection() != null) {
				clientParameters.getDataConnection().interrupt();
				clientParameters.getDataConnection().dispose();
				clientParameters.setDataConnection(null);
			}
			this.interrupt();
			if (clientParameters.getSocket() != null) {
				ArcaoFTPd.getLogger().write(("C: " + new Date().toString() + "; " + clientParameters.getSocket().getRemoteSocketAddress().toString() +": Connection closed...\n").getBytes("utf-8"));
				clientParameters.getSocket().close();
			}
			clientParameters = null;
		} catch (Exception e) {
		}
	}
}
