package cz.zcu.ups.arcaoftpd.dataaction;

import java.io.File;
import java.io.OutputStream;
import java.util.Date;
import java.util.Locale;

import cz.zcu.ups.arcaoftpd.ArcaoFTPd;

/**
 * Trida zastresujici datovou akci, ktera vylistuje adresar
 * @author Arcao
 */
public class ListDataAction extends DataAction {
	/**
	 * Pocet milisekund za den
	 */
	public static final long MILLISECS_PER_DAY =  24 * 3600 * 1000;
	/**
	 * Pocet milisekund za pul roku
	 */
	public static final long MILLISECS_PER_HALFYEAR = 182 * MILLISECS_PER_DAY;

	/**
	 * Konstriktor tridy s nastaveni parametru cesty k adresari, ktery se bude vylistovavat
	 * @param parameter			cesta
	 */
	public ListDataAction(String parameter) {
		super(parameter);
	}

	/**
	 * Uporneni klienta, ze bylo otevreno datove spojeni
	 */
	public void preRun() {
		try {
			getClientParameters().sendMessage(150, "Opening data connection for directory list.");
			ArcaoFTPd.getLogger().write(("OD: " + new Date().toString() + "; " + getClientParameters().getSocket().getRemoteSocketAddress().toString() +"; Open data connection...\n").getBytes("utf-8"));
		} catch (Exception e) {
		}
	}

	/**
	 * Vykona proces vylistovani adresare
	 */
	public void run() {
		try {
			OutputStream out = getSocket().getOutputStream();

			File dir = getClientParameters().getUser().translatePathToAbsoluteFile(getParameter());
			//pokud je to soubor, vypiseme jen informace o tomto souboru
			if (dir.isFile()) {
				System.out.print(getItemListFormat(dir));
				out.write(getItemListFormat(dir).getBytes(getClientParameters().getEncoding()));
			// jinak vypiseme vsechny polozky adresare
			} else {
				for (String item : dir.list()) {
					File itemFile = new File(dir, item);
					//System.out.print(getItemListFormat(itemFile));
					out.write(getItemListFormat(itemFile).getBytes(getClientParameters().getEncoding()));
				}

			}
			out.close();
			getSocket().close();
			getClientParameters().sendMessage(226, "Transfer ok.");
		} catch(Exception e) {
			try {
				getClientParameters().sendMessage(426, "Connection closed; transfer aborted.");
			} catch (Exception ex) {
			}
			e.printStackTrace();
		}
		try {
			ArcaoFTPd.getLogger().write(("CD: " + new Date().toString() + "; " + getClientParameters().getSocket().getRemoteSocketAddress().toString() +": Close data connection...\n").getBytes("utf-8"));
		} catch (Exception e) {
		}
	}

	/**
	 * Vrati radku s informacemi ve tvaru linuxoveho ls -l
	 * @param item						polozka jako objekt File
	 * @return							informace ve tvaru linuxoveho ls -l
	 */
	private String getItemListFormat(File item) {
		//"-rw-r--r-- 1 owner group           213 Aug 26 16:31 README.TXT";
		//"drwxr-xr-x 1 owner group           213 Aug 26 16:31 DIRECTORY";

		final long currTime = new Date().getTime();
		if (item.isFile()) {
			// pokud je soubor starsi vic jak pulrok od aktualniho data, vypiseme misto casu rok
			if (currTime - item.lastModified() < MILLISECS_PER_HALFYEAR) {
				return String.format(Locale.US, "-rw-r--r-- 1 owner group %1$ 13d %2$tb %2$td %2$tR %3$s\r\n", item.length(), new Date(item.lastModified()), item.getName());
			} else {
				return String.format(Locale.US, "-rw-r--r-- 1 owner group %1$ 13d %2$tb %2$td  %2$tY %3$s\r\n", item.length(), new Date(item.lastModified()), item.getName());
			}
		} else {
			if (currTime - item.lastModified() < MILLISECS_PER_HALFYEAR) {
				return String.format(Locale.US, "drwxr-xr-x 1 owner group %1$ 13d %2$tb %2$td %2$tR %3$s\r\n", item.length(), new Date(item.lastModified()), item.getName());
			} else {
				return String.format(Locale.US, "drwxr-xr-x 1 owner group %1$ 13d %2$tb %2$td  %2$tY %3$s\r\n", item.length(), new Date(item.lastModified()), item.getName());
			}
		}
	}
}

