package cz.zcu.ups.arcaoftpd.dataaction;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.util.Date;

import cz.zcu.ups.arcaoftpd.ArcaoFTPd;
import cz.zcu.ups.arcaoftpd.client.TransferMode;

/**
 * Trida zastresujici datovou akci, ktera posila soubor klientovi
 * @author Arcao
 */
public class RetriveDataAction extends DataAction {
	/**
	 * Velikost odesilaciho bufferu
	 */
	private static final int FILE_BUFFER_LEN = 4096;
	private File file;

	/**
	 * Konstriktor tridy s nastavenim cesty k souboru, ktery se ma poslat klientovi
	 * @param parameter		cesta k souboru
	 */
	public RetriveDataAction(String parameter) {
		super(parameter);
	}

	/**
	 * Posleme klinetovi, ze jsme otevreli spojeni
	 */
	public void preRun() {
		try {
			file = getClientParameters().getUser().translatePathToAbsoluteFile(getParameter());
			String filename = file.getName();
			if (getParameter().length() > 0 && getParameter().charAt(getParameter().length() - 1) == '/') filename = "/";
			getClientParameters().sendMessage(150, "Opening data connection for \"%s\".", filename);
			ArcaoFTPd.getLogger().write(("OD: " + new Date().toString() + "; " + getClientParameters().getSocket().getRemoteSocketAddress().toString() +"; Open data connection...\n").getBytes("utf-8"));
		} catch (Exception e) {
		}
	}

	/**
	 * Funkce starajici se o odeslani souboru klientovi
	 */
	public void run() {
		try {
			//je li spojeni v binarnim rezimu, posleme data binarne, jinak textove
			if (getClientParameters().getTransferMode() == TransferMode.Binary) {
				OutputStream out = getSocket().getOutputStream();
				InputStream in = new FileInputStream(file);

				//pokud je nastaveno, ze se ma preskocit urcity pocet bytu, preskocime je
				in.skip(getClientParameters().getFileSeekPosition());
				getClientParameters().setFileSeekPosition(0);

		        //preneseme vsechny bajty souboru (cteni se provede bufferovane o velikosti bufferu FILE_BUFFER_LEN)
		        byte[] buf = new byte[FILE_BUFFER_LEN];
		        int len;
		        while ((len = in.read(buf)) > 0) {
		            out.write(buf, 0, len);
		        }
		        in.close();
		        out.close();
			} else {
				BufferedWriter out = new BufferedWriter(new OutputStreamWriter(getSocket().getOutputStream()));
				BufferedReader in = new BufferedReader(new InputStreamReader(new FileInputStream(file)));

				String line;
				//preneseme vsechny radky v souboru
				while((line = in.readLine()) != null)
					out.write(line + "\r\n");
				in.close();
				out.close();
			}

			getSocket().close();
			getClientParameters().sendMessage(226, "Transfer ok.");
		} catch(Exception e) {
			try {
				getClientParameters().sendMessage(426, "Connection closed; transfer aborted.");
			} catch (Exception ex) {
			}
			e.printStackTrace();
		}
		try {
			ArcaoFTPd.getLogger().write(("CD: " + new Date().toString() + "; " + getClientParameters().getSocket().getRemoteSocketAddress().toString() +": Close data connection...\n").getBytes("utf-8"));
		} catch (Exception e) {
		}
	}
}
