package cz.zcu.ups.arcaoftpd.users;

import java.io.File;

/**
 * Trida zastresuje hodnoty namapovaneho adresare
 * @author SLOUP
 *
 */
public class Map implements Comparable {
	private String source;
	private String destination;
	private String access;

	/**
	 * Konstruktor objektu s urcenim pocatecnich parametru
	 * @param source			absolutni cesta k adresari, ktery chceme namapovat
	 * @param destination		cesta od korene ftp, kam se bude objekt mapovat
	 * @param access			pristupova prava k objektu
	 */
	public Map(String source, String destination, String access) {
		this.source = source;
		this.destination =  destination;
		this.access = access;

		//oprava parametru
		if (new File(this.source).isDirectory() && this.destination.charAt(this.destination.length() - 1) != '/')
			this.destination+= '/';

	}

	/**
	 * Vrati pristupova prava k objektu
	 * @return 					pristupova prava k objektu
	 */
	public String getAccess() {
		return access;
	}

	/**
	 * Nastavi pristupova prava k objektu
	 * @param access 			pristupova prava k objektu
	 * @return					referenci na (this) Map
	 */
	public Map setAccess(String access) {
		this.access = access;
		return this;
	}

	/**
	 * Vrati cestu od korene ftp, kam se bude objekt mapovat
	 * @return					cesta od korene ftp, kam se bude objekt mapovat
	 */
	public String getDestination() {
		return destination;
	}

	/**
	 * Nastavi cestu od korene ftp, kam se bude objekt mapovat
	 * @param destination		cesta od korene ftp, kam se bude objekt mapovat
	 * @return					reference na (this) Map
	 */
	public Map setDestination(String destination) {
		this.destination = destination;
		return this;
	}

	/**
	 * Vrati absolutni cestu k adresari, ktery chceme namapovat
	 * @return					absolutni cesta k adresari, ktery chceme namapovat
	 */
	public String getSource() {
		return source;
	}

	/**
	 * Nastavi absolutni cestu k adresari, ktery chceme namapovat
	 * @param source			absolutni cesta k adresari, ktery chceme namapovat
	 * @return					referenci na (this) Map
	 */
	public Map setSource(String source) {
		this.source = source;
		return this;
	}

	/**
	 * Porovnava dva objekty tridy Map
	 * @param o					objekt, s kterym se bude srovnavat
	 * @return					-1, 0 nebo 1, pokud je zadany objekt vestsi, mensi nebo roven
	 */
	public int compareTo(Object o) {
		if (o instanceof Map) {
			if (((Map) o).getDestination().compareToIgnoreCase(getDestination()) != 0)
				return ((Map) o).getDestination().compareToIgnoreCase(getDestination());
			if (((Map) o).getDestination().compareToIgnoreCase(getSource()) != 0)
				return ((Map) o).getDestination().compareToIgnoreCase(getSource());
			return 0;
		} else {
			return o.toString().compareTo(toString());
		}
	}

	/**
	 * Ma namapovany objekt prava pro cteni
	 * @return					pravo pro cteni
	 */
	public boolean isReadable() {
		return (access.indexOf('r') != -1);
	}

	/**
	 * Ma namapovany objekt prava pro zapis
	 * @return					pravo pro zapis
	 */
	public boolean isWriteable() {
		return (access.indexOf('w') != -1);
	}

	/**
	 * Ma namapovany objekt prava pro vylistovani sveho obsahu
	 * @return					prvavo pro vylistovani sveho obsahu
	 */
	public boolean isListable() {
		return (access.indexOf('l') != -1);
	}

	/**
	 * Ma namapovany objekt pravo smazat sve soubory ci sebe (soubor)
	 * @return					pravo se smazat sve soubory ci sebe (soubor)
	 */
	public boolean isFilesDeleteable() {
		return (access.indexOf('u') != -1);
	}

	/**
	 * Ma namapovany objekt pravo smazat sve adresare ci sebe (adresar)
	 * @return					pravo se smazat sve adresare ci sebe (adresar)
	 */
	public boolean isDirectoryDeleteable() {
		return (access.indexOf('d') != -1);
	}

	/**
	 * Ma namapovany objekt pravo navazat dalsi data na svuj konec (soubor)
	 * @return					pravo navazat dalsi data na svuj konec (soubor)
	 */
	public boolean isAppendable() {
		return (access.indexOf('a') != -1);
	}

	/**
	 * Ma namapovany objekt pravo vytvaret podadresare
	 * @return					pravo vytvaret podadresare
	 */
	public boolean isDirectoryMakeable() {
		return (access.indexOf('m') != -1);
	}

	/**
	 * Je namapovany objekt adresar
	 * @return					je to adresar
	 */
	public boolean isDirectory() {
		return new File(this.source).isDirectory();
	}

	/**
	 * Je namapovany objekt soubor
	 * @return					je to soubor
	 */
	public boolean isFile() {
		return new File(this.source).isFile();
	}
}

